#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 27 02:37:12 2017

Default plotting parameters

@author: Pawel M. Kozlowski
"""

# importing Python modules
import numpy as np
import matplotlib 
import matplotlib.pyplot as plt
from itertools import cycle


# listing all functions declared in this file so that sphinx-automodapi
# correctly documents them and doesn't document imported functions.
__all__ = ["plot_line_shaded",
           "plot_scatter_bars"]


# setting default plot properties
plotFont = {'family' : 'serif',
            'weight' : 'normal',
            'size'   : 16}
plt.rc('font', **plotFont)
plt.rc('lines', linewidth=2)
plt.rc('lines', markersize=8)
#plt.rc('text', usetex=True)
#plt.rc('text', usetex=False)
# setting default figure size
matplotlib.rcParams['figure.figsize'] = 10, 6

validMathTextFonts = ['dejavusans',
                      'dejavuserif',
                      'cm',
                      'stix',
                      'stixsans',
                      'custom']

# setting latex rendered fonts to be same as regular fonts
try:
    matplotlib.rcParams['mathtext.fontset'] = 'dejavuserif'
    matplotlib.rcParams['mathtext.rm'] = 'DejaVu Serif'
    matplotlib.rcParams['mathtext.it'] = 'DejaVu Serif:italic'
    matplotlib.rcParams['mathtext.bf'] = 'DejaVu Serif:bold'
except:
    print("Couldn't load dejavuserif fonts for plot defaults."
          "Falling back to stix fonts.")
    try:
        matplotlib.rcParams['mathtext.fontset'] = 'stix'
        matplotlib.rcParams['font.family'] = 'STIXGeneral'
    except:
        print("Couldn't load stix fonts for plot defaults.")

# for testing whether matplotlib and python fonts match
#plt.title(r'cm$\rm cm^{-3}$')


#%% convenience functions for plotting

# shaded error bars for line plot
def plot_line_shaded(xData, yData, yErrsPos, yErrsNeg=[], label="", **kwargs):
    """
    Generate a line plot with shaded region representing y-error bars.
    Can be run multiple times before plt.show(), to plot multiple data
    sets on the same axes.
    
    Parameters
    ----------
    xData: numpy.ndarray
        X-axis data to be plotted.
        
    yData: numpy.ndarray
        Y-axis data to be plotted.
        
    yErrsPos: numpy.ndarray
        Errors on yData.
        
    yErrsPos: numpy.ndarray
        When errors on yData are asymmetric, these are the positive
        side errors.
        
    yErrsNeg: numpy.ndarray
        When errors on yData are asymmetric, these are the negative
        side errors.
    
    Returns
    -------
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    #make errors symmetric unless specified otherwise
    if yErrsNeg == []:
        yErrsNeg = yErrsPos
    # check that arrays are 1D
    
    # check that arrays are equal length
    if not len(xData) == len(yData) == len(yErrsPos) ==len(yErrsNeg):
        raise ValueError("Arrays must of equal length!")

    # line plot
    ax = kwargs.pop('ax', plt.gca())
    base_line, = ax.plot(xData, yData, label=label, **kwargs)
    # shaded error region
    plt.fill_between(xData,
                     yData - yErrsNeg,
                     yData + yErrsPos,
                     alpha=0.5,
                     edgecolor=base_line.get_color(),
                     facecolor=base_line.get_color())
    return


def plot_scatter_bars(xData, yData, yErrsPos, yErrsNeg=[], label="", **kwargs):
    """
    Generate a scatter plot with y-error bars.
    Can be run multiple times before plt.show(), to plot multiple data
    sets on the same axes.
    
    Parameters
    ----------
    xData: numpy.ndarray
        X-axis data to be plotted.
        
    yData: numpy.ndarray
        Y-axis data to be plotted.
        
    yErrsPos: numpy.ndarray
        Errors on yData.
        
    yErrsPos: numpy.ndarray
        When errors on yData are asymmetric, these are the positive
        side errors.
        
    yErrsNeg: numpy.ndarray
        When errors on yData are asymmetric, these are the negative
        side errors.
    
    Returns
    -------
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    #make errors symmetric unless specified otherwise
    if yErrsNeg == []:
        yErrsNeg = yErrsPos
    # check that arrays are 1D
    
    # check that arrays are equal length
    if not len(xData) == len(yData) == len(yErrsPos) == len(yErrsNeg):
        raise ValueError("Arrays must of equal length!")
        
    # line plot
    ax = kwargs.pop('ax', plt.gca())
    ax.errorbar(xData,
                yData,
                yerr=[yErrsNeg, yErrsPos],
                label=label,
                fmt='o',
                fillstyle='none',
                capsize=4,
                elinewidth=1,
                **kwargs)
    return

#%% testing custom plotting tools

#xData = np.arange(10)
#yData = xData ** 2
#yData2 = xData ** 3
#yErrs = np.ones_like(xData) * 30
#plot_line_shaded(xData, yData, yErrs, label="#1", color='r')
#plot_line_shaded(xData, yData2, yErrs, label="#2")
#plt.ylabel("y")
#plt.xlabel("x")
#plt.legend(loc='upper left', frameon=False, labelspacing=0.001,
#           fontsize=14, borderaxespad=0.4)
#plt.show()
#
#
#xData = np.arange(10)
#yData = xData ** 2
#yData2 = xData ** 3
#yErrs = np.ones_like(xData) * 30
#plot_scatter_bars(xData, yData, yErrs, label="#1", color='r')
#plot_scatter_bars(xData, yData2, yErrs, label="#2")
#plt.ylabel("y")
#plt.xlabel("x")
#plt.legend(loc='upper left', frameon=False, labelspacing=0.001,
#           fontsize=14, borderaxespad=0.4)
#plt.show()

#%% style cycling
# printing default colors
default_colors = matplotlib.colors.cnames.keys()
#print(f"Colors: {default_colors}")
# printing default linestyles
default_lines = matplotlib.lines.lineStyles.keys()
#print(f"Lines: {default_lines}")
# printing default marker styles
default_markers = matplotlib.markers.MarkerStyle.markers.keys()
#print(f"Markers: {default_markers}")

# custom list of linestyles (excluding blank line styles)
lines = ["-","--","-.",":"]
linecycler = cycle(lines)
#plt.figure()
#for i in range(10):
#    x = range(i,i+10)
#    plt.plot(range(10),x,next(linecycler))
#plt.show()
    

#%% Custom colors based on https://towardsdatascience.com/making-matplotlib-beautiful-by-default-d0d41e3534fd
#CB91_Blue = '#2CBDFE'
#CB91_Green = '#47DBCD'
#CB91_Pink = '#F3A0F2'
#CB91_Purple = '#9D2EC5'
#CB91_Violet = '#661D98'
#CB91_Amber = '#F5B14C'
#
#color_list = [CB91_Blue,
#              CB91_Pink,
#              CB91_Green,
#              CB91_Amber,
#              CB91_Purple,
#              CB91_Violet]
#plt.rcParams['axes.prop_cycle'] = plt.cycler(color="seaborn-colorblind")


#%% built-in custom color cycles
#print(plt.style.available)
#plt.style.use('seaborn')


# -*- coding: utf-8 -*-
"""
Created on Tue Jul 21 12:16:31 2020

Testing for stats.py

@author: Myles T. Brophy
"""

# python modules
import numpy as np
import pytest

# custom modules
import fiducia.stats as stats

class Test_integrationError(object):
    def setup_class(self):
        """Initialize parameters for tests."""
        self.rectangle = np.full(2,5)
        self.rectangleHeightUncertainty = np.full(2, 1) 
        
        self.diagonal = np.arange(5)*3 + 1
    
    def test_trapz(self):
        #check that the trap integration of a rectangle equals width*height
        assert np.trapz(self.rectangle) == (len(self.rectangle) -1)*self.rectangle[0]
        #check that the trap integration of a diagonal equals the area of a right trapezoid.
        assert np.trapz(self.diagonal) == 0.5 * (len(self.diagonal)-1) * (self.diagonal[0] + self.diagonal[-1])
    
    def test_trapzUncertainty(self):
        """Testing for expected return value."""
        assert stats.trapzVariance([1,2,3]) == 6.5
        assert stats.trapzVariance([1,2,3], x=[5,7,10]) == 46.25
        assert stats.trapzVariance([1,2], x=[1,3]) == 5.0 
        assert stats.trapzVariance([1], x=[2]) == 0.0
        #assert misc.trapzVariance(self.rectangleHeightUncertainty) == self.rectangleHeightUncertainty[0]*(len(self.rectangleHeightUncertainty)-1)
    
    def test_interpVariance(self):
        """Testing for expected return value."""
        #checked by hand
        assert np.array_equal(stats.interpVariance([1,4,6], [1,2,3,4,5,6,7,8], [2,4,3,5,2,5,1,2]),[4.0, 25.0,25.0])
        assert np.array_equal(stats.interpVariance([0,4,4.5,6,9], [1,2,3,4,5,6,7,8], [2,4,3,5,2,5,1,2]),[4.0, 25.0, 7.25,25.0,4.0])


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  8 09:25:23 2019

Tests for misc.py

@author: Pawel M. Kozlowski, Myles T. Brophy
"""

# python modules
import numpy as np
import pytest

# custom modules
import fiducia.misc as misc


def test_find_nearest():
    """
    Testing for expected return value.
    """
    ex = [1,10,100,500,1000];
    assert misc.find_nearest(ex, 40) == (1,10)
    assert misc.find_nearest(ex, 5000) == (4,1000)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 18:10:57 2020

Place for setting up fixtures to be used across pytests.
Loading example dante datasets should be done here so that they
don't have to be repeated across the various test modules.

@author: Pawel M. Kozlowski
"""

import pytest
import numpy as np


@pytest.fixture(scope="session")
def example_info(tmpdir_factory):
    r"""
    Loads example information on Dante data (calibration files, and raw data)
    from Omega-60, which can be used across multiple pytests.
    """
    # response function curves for Dante channels.
    responseFile = 'example_response.csv'
    # attenuator information loopkup by ID number.
    attenuatorsFile = 'TableAttenuators.xls'
    # Dante oscilloscope voltage axis offsets.
    offsetsFile = 'Offset.xls'
    # Raw Dante data file from Omega-60.
    dataFile = 'dante_dante86455.dat'
    # emitting area of blackbody in mm^2.
    area = np.pi * 0.6 ** 2
    # dante viewing angle relative to normal of emitting surface, in degrees.
    angle = 69.09
    # collecting info into a dict which can be passed to various pytests.
    info = {"Response File" : responseFile,
            "Attenuators File" : attenuatorsFile,
            "Offsets File" : offsetsFile,
            "Data File" : dataFile,
            "Area" : area,
            "Angle" : angle}
    return info



@pytest.fixture(scope="session")
def load_example_data(example_info):
    r"""
    Loads the example Dante data from example_info() so that it only
    has to be done once across the session.
    """
    # load raw dante data into corrected/aligned format
    from fiducia.rawProcess import loadCorrected
    loaded = loadCorrected(danteFile=example_info["Data File"],
                           attenuatorsFile=example_info["Attenuators File"],
                           offsetsFile=example_info["Offsets File"],
                           plot=False)
    timesFrame, dfAtten, onChList, hf, dfVolt = loaded
    dataDict = {"Times Frame" : timesFrame,
                "Attenuators Frame" : dfAtten,
                "On Channels" : onChList,
                "Header Frame" : hf,
                "Signal Frame" : dfVolt}
    return dataDict


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jan 25 12:22:22 2019

Tests for main.py

Tests should be split and marked as:
    -type check tests, which check function execution and error handling
    against inputs of different types.
    - numerical tests, which check whether given input values produce
    expected output values.
    - property tests, which check mathematical properties of functions,
    such as associativity, commutativity, identity, etc.
    - physical tests, which check physical intution, such as higher power
    should produce larger radiation temperature, or larger uncertainties
    on the inputs should produce larger uncertainties on the outputs.


@author: Pawel M. Kozlowski, Myles T. Brophy
"""

# python modules
import numpy as np
import pytest
from hypothesis import given
import hypothesis.strategies as s
import hypothesis.extra.numpy as hnp
import hypothesis.extra.pandas as hpd

# custom modules
import fiducia.main as main


# categorize this test as part of the group of tests which check
# expected values of functions.
@pytest.mark.value
# parametrize the test to run multiple times over different inputs.
@pytest.mark.parametrize("power,area,angle,expectedTRad",
                         [(0, 0.6, 30, 0),
                          (1, 0.6, 30, 36.98703164858843),
                          (10, 0.6, 30, 65.7732768191428),
                          (100, 0.6, 30, 116.963263898272)])
def test_inferRadTemp_val(power, area, angle, expectedTRad):
    """Testing for expected return value."""
    methodVals, methodValsVariance = main.inferRadTemp(power, 
                                                       area,
                                                       angle)
    testTrue = np.allclose(methodVals,
                           expectedTRad, 
                           rtol=1e-7,
                           atol=0.0)
    errStr = (f"Infer radiation temperture test gives {methodVals} but "
              f"should be {expectedTRad}")
    assert testTrue, errStr
        

@given(power=s.floats(min_value=0),
       powerUncertainty=s.floats(min_value=0))
def test_inferRadTemp_prop(power, powerUncertainty):
    """
    Property-based testing of inferRadTemp() using hypothesis.
    Given a physical value of power (>=0),
    expect radiation temperature to be phyiscal (>=0).
    """
    area = 0.6
    angle = 30
    methodVals, methodValsVariance = main.inferRadTemp(power=power, 
                                                       area=area,
                                                       angle=angle,
                                                       powerUncertainty=powerUncertainty)
    assert methodVals >= 0 and methodValsVariance >= 0


"""
Created on Fri Mar  8 09:41:36 2019

Functions for working with cubic spline equation in matrix form.

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np
import scipy.sparse as sparse
from scipy import integrate
import matplotlib.pyplot as plt
import xarray as xr
# custom modules
import fiducia.pltDefaults
from fiducia.stats import dotVariance

# listing all functions declared in this file so that sphinx-automodapi
# correctly documents them and doesn't document imported functions.
__all__ = ["splineCoords",
           "splineCoordsInv",
           "yCoeffArr",
           "dCoeffArr",
           "dToyArr",
           "responseInterp",
           "yChiCoeffArr",
           "yChiCoeffArrEnergies",
           "fancyTrapz2",
           "segmentsArr",
           "detectorArr",
           "knotSolve",
           "reconstructSpectrum",
           ]


def splineCoords(energy, energyStart, energyEnd):
    r"""
    Convert photon energy value into normalized coordinates for a particular
    spline region.
    
    Parameters
    ----------
    energy: float, numpy.ndarray
        Energy value to be converted into normalized spline coordinate.
        
    energyStart: float
        Lower bound energy for the spline region based on knot points.
        
    energyEnd: float
        Upper bound energy for the spline region based on knot points.
    
    normCoord: float
        Return value of energy converted into normalized spline coordinates.
    
    Returns
    -------
    normCoord: float, numpy.ndarray
        Normalized energy coordinate(s).
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    
    """
#    if energyStart > energyEnd:
#        raise ValueError(f"Lower bound energy {energyStart} must be less than "
#                        f"upper bound energy {energyEnd}.")
#    if np.all(energy < energyStart) or np.all(energy > energyEnd):
#        raise ValueError(f"Given energy value {energy} is outside spline "
#                        f"coordinate bounds {energyStart} to {energyEnd}.")
    normCoord = (energy - energyStart) / (energyEnd - energyStart)
    return normCoord


def splineCoordsInv(energyNorm, energyStart, energyEnd):
    r"""
    Given a normalized energy value and the bounds of a spline segment, 
    return the un-normalized photon energy value. This is the inverse of
    splineCoords().
    
    Parameters
    ----------
    energyNorm: float, numpy.ndarray
        Normalized photon energy
        
    energyStart: float
        Lower bound energy for the spline region based on knot points.
        
    energyEnd: float
        Upper bound energy for the spline region based on knot points.
    
    Returns
    -------
    energy: float, numpy.ndarray
        Absolute photon energy (un-normalized).
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    energy = (energyEnd - energyStart) * energyNorm + energyStart
    return energy


def yCoeffArr(energyNorm, chLen):
    r"""
    Returns the matrix M_y(t) for a given value of t in:
        
    .. math::
        Y_i(t) = M_y(t) y_i + M_D(t) D_i
        
    Parameters
    ----------
    energyNorm: float
        normalized photon energy for a spline section.
        
    chLen: int
        Number of DANTE channels (equal to number of spline knots).
    
    Returns
    -------
    mArr: scipy.sparse.lil.lil_matrix
        Sparse matrix :math:`M_y(t)`.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    coeff1 = 1 - 3 * energyNorm ** 2 + 2 * energyNorm **3 
    coeff2 = 3 * energyNorm ** 2 - 2 * energyNorm ** 3
    mArr = sparse.diags([coeff1, coeff2],
                        [0,1],
                        shape=(chLen, chLen + 1),
                        format='lil')
    #First and last row elements in y array are different (added by DHB 3/25/19)
    mArr[0,0] = 1 - energyNorm ** 3
    mArr[0,1] = energyNorm ** 3
    mArr[chLen-1,chLen-1] = 1 - energyNorm ** 3
    mArr[chLen-1,chLen] = energyNorm ** 3
    return mArr.tocsc()


def dCoeffArr(energyNorm, chLen):
    r"""
    Returns the matrix M_D(t) for a given value of t in:
        
    .. math::
        Y_i(t) = M_y(t) y_i + M_D(t) D_i
    
    Parameters
    ----------
    energyNorm: float
        normalized photon energy for a spline section.
        
    chLen: int
        Number of DANTE channels (equal to number of spline knots).
    
    Returns
    -------
    dArr: scipy.sparse.lil.lil_matrix
        Sparse matrix :math:`M_D(t)`.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    coeff1 = energyNorm - 2 * energyNorm ** 2 + energyNorm ** 3
    coeff2 = -1 * energyNorm ** 2 + energyNorm ** 3
    dArr = sparse.diags([coeff1, coeff2],
                        [0,1],
                        shape=(chLen, chLen + 1), 
                        format='lil')
    #First and last row elements in D array are different (added by DHB 3/25/19)
    dArr[0,0] = energyNorm - energyNorm ** 3
    dArr[chLen-1,chLen-1] = energyNorm - energyNorm ** 2
    dArr[chLen-1,chLen] = energyNorm ** 2 - energyNorm ** 3
    return dArr.tocsc()


def __chi1Arr__(chLen):
    r"""
    Parameters
    ----------
    chLen: int
        Number of DANTE channels (equal to number of spline knots).
    
    Returns
    -------
    chi1: scipy.sparse.lil.lil_matrix
        Sparse matrix :math:`\chi_1`.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # chi1array in Dan Barnak's notation = TDy'' in Jim's paper 
    chi1 = sparse.diags([1, 4, 1], 
                        [-1,0,1],
                        shape=(chLen + 1, chLen + 1),
                        format='lil')
    chi1[0, 0] = 2
    chi1[-1, -1] = 2
    return chi1.tocsc()


def __chi3Arr__(chLen):
    r"""
    Parameters
    ----------
    chLen: int
        Number of DANTE channels (equal to number of spline knots).
        
    Returns
    -------
    chi3: scipy.sparse.lil.lil_matrix
        Sparse matrix :math:`\chi_3`.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # chi2array in Dan Barnak's notation = TDy in Jim's paper
    chi3 = sparse.diags([-1, 0, 1],
                        [-1,0,1],
                        shape=(chLen + 1, chLen + 1),
                        format='lil')
    chi3[0, 0] = -1
    chi3[-1, -1] = 1
    return chi3.tocsc()


def dToyArr(chLen):
    r"""
    Construct matrix for converting from :math:`D_i` to :math:`y_i` vector.
    
    Parameters
    ----------
    chLen: int
        Number of DANTE channels (equal to number of spline knots).
        
    Returns
    -------
    diToyi: numpy.ndarray
        Matrix for converting from :math:`D_i` to :math:`y_i` vector.
        
    Notes
    -----
    The matrix is given by:
        
    .. math::
        D_i = 3 \chi_1^{-1} \chi_3 y_i
    
    
    See also
    --------
    
    Examples
    --------
    """
    chi1Arr = __chi1Arr__(chLen)
    chi3Arr = __chi3Arr__(chLen)
    chi1ArrInv = np.linalg.inv(chi1Arr.toarray())
    # constructing array for converting values from Di to yi
    diToyi = 3 * np.dot(chi1ArrInv, chi3Arr.toarray())
    return diToyi


def responseInterp(energyNorm, energyMin, energyMax, responseFrame, channels):
    r"""
    Given a DANTE detector response as a function of energy, convert the
    response to normalized photon energy, t, over a given spline segment, and
    return interpolated response values for a given value of t. Returns an
    array of interpolated responses corresponding to the number of channels.
    
    Parameters
    ----------
    energyNorm: float, numpy.ndarray
        normalized photon energy
        
    energyMin: float
        Lower bound photon energy of the spline segment over which we are
        normalizing.
        
    eneryMax: float
        Upper bound photon energy of the spline segment over which we are
        normalizing.
        
    responseFrame: pandas.core.frame.DataFrame
        DANTE channel responses as a function of photon energy (not normalized).

    channels: numpy.ndarray
        numpy array of DANTE channel numbers.
    
    Returns
    -------
    responsesInterpd: numpy.ndarray
        Returns a matrix of (energyNorms, channels) of response functions.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------

    """
    chLen = len(channels)
    if np.shape(energyNorm):
        # if energyNorm is a vector, then get the length of that vector
        energyLen = len(energyNorm)
    else:
        # otherwise the length is just 1
        energyLen = 1
    # construct an array of interpolated response functions based on
    # number of photon energy points we want and number of DANTE channels
    responsesInterpd = np.zeros((energyLen, chLen))
    # fetch the original array of energy values (not normalized)
    energyArr = responseFrame['Energy(eV)']
    # converting normalized energy values to un-normalized energy values,
    # since our response functions are in terms of absolute photon energy
    energyReg = splineCoordsInv(energyNorm, energyMin, energyMax)
    for idx, channel in enumerate(channels):
        responseArr = responseFrame[channel]
        responsesInterpd[:,idx] = np.interp(x=energyReg,
                                            xp=energyArr,
                                            fp=responseArr)
    return responsesInterpd


def yChiCoeffArr(energyNorm, chLen, dToY):
    r"""
    This is the matrix corresponding to:
        
    .. math::
        M_y(t) + 3 M_D(t) \chi_1^{-1} \chi_3
    
    which describes the cubic spline interpolation of the x-ray
    spectrum.
    
    Parameters
    ----------
    energyNorm: float
        normalized photon energy
        
    chLen: int
        Number of DANTE channels (equal to number of spline knots).
        
    dToY: numpy.ndarray
        Matrix for converting from :math:`D_i` to :math:`y_i` values in
        cubic spline interpolation. See dToyArr().
    
    Returns
    -------
    yChiArr: numpy.ndarray
        Returns a 2D matrix for a particular value of energyNorm.
        
    Notes
    -----
    The matrix is given by:
        
    .. math::
        M_y(t) + 3 M_D(t) \chi_1^{-1} \chi_3
    
    See also
    --------
    
    Examples
    --------
    """
    # M_y(t)
    yCoeff = yCoeffArr(energyNorm, chLen)
    # M_D(t)
    dCoeff = dCoeffArr(energyNorm, chLen)
    # folded x-ray spectrum and detector response array
    yChiArr = yCoeff.toarray() + np.dot(dCoeff.toarray(), dToY)
    return yChiArr


def yChiCoeffArrEnergies(energyNorms, chLen, dToY):
    r"""
    This is the matrix corresponding to:
        
    .. math::
        M_y(t) + 3 M_D(t) \chi_1^{-1} \chi_3
    
    which describes the cubic spline interpolation of the x-ray
    spectrum.
    
    energyNorms: numpy.ndarray
        Vector of normalized photon energies.
        
    chLen: int
        Number of DANTE channels (equal to number of spline knots).
        
    dToY: numpy.ndarray
        Matrix for converting from :math:`D_i` to :math:`y_i` values in cubic
        spline interpolation. See dToyArr().
    
    Returns
    -------
    yChiArrEnergies: numpy.ndarray
        Returns a 3D matrix composed of a series of 2D yChiCoeff matrices
        corresponding to the given energyNorm values. This matrix is
        indexed as (energyNorms, knotIndex, knotIndex).
        
    Notes
    -----
    The matrix is given by:
        
    .. math::
        M_y(t) + 3 M_D(t) \chi_1^{-1} \chi_3
    
    See also
    --------
    
    Examples
    --------
    """
    yChiArrEnergies = np.array([yChiCoeffArr(energyNorm, chLen, dToY) for energyNorm in energyNorms])
    return yChiArrEnergies
    

def fancyTrapz2(energyNorms, yChis, segments, responseFrame, channels):
    r"""
    Trap rule integration of the folding between our :math:`M_{y \chi}` matrix
    and response function matrix, with respect to normalized photon energy,
    for each channel. The result should be a matrix with shape
    (`len(channels)`, `len(segments)`, `len(knotIndex)`).
    
    Parameters
    ----------
    energyNorms: numpy.ndarray
        1D array of normalized photon energies
        
    yChis: numpy.ndarray
        3D array of :math:`M_{y \chi}` values corresponding to
        (energyNorms, segments, knotIndex).
        
    responses: numpy.ndarray
        2D array of DANTE channel response functions corresponding to
        (energyNorms, channels).
        
    channels: numpy.ndarray
        Array of DANTE channel numbers.      
        
    Returns
    -------
    integArr : xarray.DataArray
        A matrix containing the folded integration of the :math:`M_{y \chi}` 
        matrix and response function matrix, with respect to normalized photon 
        energy. Has shape (`len(channels)`, `len(segments)`, `len(knotIndex)`).
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    shape = np.shape(yChis)
    segmentsLen = np.shape(segments)[0]
    knotsLen = shape[2]
    chLen = segmentsLen
    # initialize integArr for storing photon energy integrated values
    integArr = xr.DataArray(np.zeros((chLen, segmentsLen, knotsLen)),
                            dims=['channel', 'segment', 'knot_point'],
                            coords={'channel':channels})
    # loop over relevant DANTE channels for analysis
    for channelIdx in np.arange(integArr['channel'].size):
        # loop over photon energy segments (between knot points)
#        print(f"Calculating for channel {channels[channelIdx]}.")
        for segmentNum, segment in enumerate(segments):
            energyMin, energyMax = segment
#            print(f"Calculating for segment {segments[segmentNum]}.")
            # loop over knot points
            for knotNum in np.arange(integArr['knot_point'].size):
                # multiplication of response by spline matrix
                responses = responseInterp(energyNorms,
                                           energyMin,
                                           energyMax,
                                           responseFrame,
                                           channels)
#                plt.plot(energyNorms, responses[:, channelIdx])
#                plt.title(f"Response channel {channels[channelIdx]}, segment {segmentNum}, knot {knotNum}")
#                plt.show()
                multArr = yChis[:, segmentNum, knotNum] * responses[:, channelIdx]
#                integArr[channelIdx, segmentNum, knotNum] = np.trapz(y=multArr,
#                                                                     x=energyNorms)
                integVal = (energyMax - energyMin) * integrate.simps(y=multArr, x=energyNorms)
                integArr[channelIdx, segmentNum, knotNum] = integVal
    return integArr


def segmentsArr(knots):
    r"""
    Returns the bounds of each spline segment, given the spline knot points.
    
    Returns an array of tuples of (energyMin, energyMax) describing the bounds
    of each spline segment, given an array of spline knots (photon energies
    corresponding to K-edges).
    
    Parameters
    ----------
    knots: numpy.ndarray
        numpy array of photon energies describing positions of spline knots.
        
    Returns
    -------
    segments : numpy.ndarray
        A 1D array of tuples of (energyMin, energyMax), corresponding to the 
        bounds of each spline segment. Has length of 'len(knots) - 1'.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    # initializing array of segments
    segments = np.zeros(len(knots) - 1, dtype=object)
    for idx, knot in enumerate(knots):
        if idx == len(segments):
            # skipping the last knot since there are 1 fewer segments than knots
            continue
        segments[idx] = knots[idx], knots[idx + 1]
    return segments
    

def detectorArr(channels, knots, responseFrame, boundary="y0", npts=1000):
    r"""
    Matrix representing the spectrally integrated folding of the detector
    response with a cubic spline interpolation of the x-ray spectrum. This
    is applied to the measured DANTE channel signals to recover knot points
    :math:`y_i` of the cubic spline, which can then be used to reconstruct the
    inferred x-ray spectrum.
    
    Parameters
    ----------
    channels: numpy.ndarray
        Array of DANTE channel numbers.
    
    knots: numpy.ndarray
        Array of photon energies describing positions of spline knots.
    
    responseFrame: pandas.core.frame.DataFrame
        DANTE channel responses as a function of photon energy (not 
        normalized).
    
    boundary: str, optional
        Choose whether yGuess corresponds to :math:`y_0` (lowest photon 
        energy) or :math:`y_{n+1}` (highest photon energy) boundary condition.
        This should correspond to the photon energy value given in knots.
        Options are 'y0' or 'yn+1'. Default 'y0'.
        
    npts: int
        Number of points used in computing the integral
        
    Returns
    -------
    detArr: numpy.ndarray
        Matrix representing the spectrally integrated folding of the detector
        response with a cubic spline interpolation of the x-ray spectrum.
        2D array of channels and knot points of shape (n, n).
        
    Notes
    -----
    For each spline segment we have:
        
    .. math::
        M_{stuff} = \int_0^{1} (M_y(t) + 3 M_D(t) \chi_1^{-1} \chi_3) R_d(t) dt
    
    Each spline is then summed to form the full detector matrix for recovering
    the knot points.
    
    See also
    --------
    
    Examples
    --------
    """
    # number of DANTE channels where we have useful measurements
    chLen = len(channels)
    # initialize normalized energies array
    # array of normalized energies over which we do the integral
    energyNorms = np.linspace(0, 1, num=npts)
    # producing segments from knots
    segments = segmentsArr(knots)
    # calculating array for converting from values of D_i to y_i. This
    # is an optimization as this array is constant!
    dToY = dToyArr(chLen)
    # M_{y \chi} coefficients array corresponding to given normalized
    # energies. Array shape is (energyNorms, segments, knotIndex).
    yChis = yChiCoeffArrEnergies(energyNorms, chLen, dToY)

    integFoldArr = fancyTrapz2(energyNorms,
                               yChis,
                               segments,
                               responseFrame,
                               channels)
    # sum along segment axis, as each segment must contribute to the
    # overall signal.
    #detArr = np.sum(integFoldArr, axis=1)
    detArr = integFoldArr.sum(dim="segment")
    detArr.attrs['boundary'] = boundary
    if boundary == "y0":
        # extracting column corresponding to y0
        detArrBoundaryCol = detArr.isel(knot_point=0)
        detArr = detArr.isel(knot_point=slice(1, None))
    elif boundary == "yn+1":
        # extracting column corresponding to y_{n+1}
        detArrBoundaryCol = detArr.isel(knot_point=-1)
        detArr = detArr.isel(knot_point=slice(None, -1))
    else:
        raise Exception(f"No method found for boundary {boundary}.")
    return detArr, detArrBoundaryCol


def knotSolve(signals,
              detArr,
              detArrBoundaryCol,
              detArrVarianceBoundaryCol,
              detArrInv,
              stdDetArrInv,
              signalsUncertainty=None,
              yGuess=1e-77,
              npts=1000):
    r"""
    Get knot points :math:`y_i` from measured DANTE signals :math:`S_d`.
    
    Parameters
    ----------
    signals: numpy.ndarray
        numpy array of DANTE measured signal for each channel at a particular
        point in time.
        
    detArr : numpy.ndarray
        Matrix representing the spectrally integrated folding of the detector
        response with a cubic spline interpolation of the x-ray spectrum.
        2D array of channels and knot points of shape (n, n).
    
    detArrBoundaryCol : xarray.DataArray
        Column of cublic spline matrix corresponding to the knots at the 
        boundary chosen with `boundary`.
    
    detArrVarianceBoundaryCol: xarray.DataArray
        Column of variances in the cublic spline matrix corresponding to the 
        knots at the boundary chosen with `boundary`.
        
    detArrInv : xarray.DataArray
        Inversion of detArr, with the column corresponding to boundary removed 
        so detArr is invertible.
    
    stdDetArrInv: xarray.DataArray
        Array of the standard deviation of each element in detArrInv based on
        variance using the 'responseUncertaintyFrame' propagated with Monte Carlo. 

    signalsUncertainty: xarray.DataArray, optional
        numpy array of the uncertainty of the DANTE measured signal for each 
        channel at a particular point in time. The default is None.

    yGuess: float, optional
        Guess for position of boundary knot point. Default is 1e-77.
    
    npts: int, optional
        Number of points used in computing the integral. Default is 1000.
        
    Returns
    -------
    knotsY : numpy.ndarray
        Array of knot point intensity values with yGuess appended.
        
    knotsYVariance: numpy.ndarray
        Array with each element corresponding to the variance of the same
        element in 'knotsY'.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    
    """ 
    #subtract boundary col from signals
    signalsy0 = signals - yGuess * detArrBoundaryCol
    signalsy0Variance = signalsUncertainty**2  + yGuess**2 * detArrVarianceBoundaryCol
    signalsy0Uncertainty = np.sqrt(signalsy0Variance)
    # applying inverted array to signals to recover knot points
    #huge difference between xarray.dot and np.dot. See issue in Gitlab
    knotsY = np.dot(detArrInv, signalsy0)
    #knotsYother = detArrInv.dot(signalsy0, {subscripts:'i,i'})
    #print("knot diff", knotsYother-knotsY)
    knotsYVariance = dotVariance(detArrInv, signalsy0, stdDetArrInv, signalsy0Uncertainty)
    return knotsY, knotsYVariance


def reconstructSpectrum(chLen,
                        knots,
                        knotsY,
                        knotsYUncertainty=None,
                        npts=1000,
                        plot=False):
    r"""
    Reconstruct the inferred DANTE spectrum given the knot points
    :math:`y_i` obtained from knotSolve().
    
    Parameters
    ----------
    chLen: int
        Number of DANTE channels (equal to number of spline knots).
        
    knots: list, numpy.ndarray
        List or array of knot point photon energy value. See knotFind().
        
    knotsY: numpy.ndarray
        Array of knot point intensity values with yGuess appended.
        See knotSolve() and analyzeSpectrum().

    knotsYUncertainty: numpy.ndarray
        Array of knot point intensity uncertainty values with yGuess appended.
        See knotSolve() and analyzeSpectrum(). The default is None.
    
    npts: int
        Number of points used in computing the integral. The default is 1000.
        
    plot: Bool
        Flag for plotting unfolded spectrum. The default is False.
        
    Returns
    -------
    photonEnergies: numpy.ndarray
        Photon energy axis of unfolded spectrum.
        
    intensities: numpy.ndarray
        Spectral intensity axis of unfolded spectrum.
        
    intensitiesVariance: numpy.ndarray
        Uncertaitny (1 :math:`\sigma`) on spectral intensity values.
    
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    
    """
    #check for no uncertainty provided.
    if knotsYUncertainty is None:
        #give them uncertainty of 0
        knotsYUncertainty = np.zeros(knotsY.shape)
    # initialize normalized energies array
    energyNorms = np.linspace(0, 1, num=npts)
    # producing segments from knots
    segments = segmentsArr(knots)
    # getting cubic spline matrices for reconstructing spline
    dToY = dToyArr(chLen)
    yChis = yChiCoeffArrEnergies(energyNorms, chLen, dToY)
    #set a constant simple uncertainty for the yChis.
    #TOOD replace with something physical
    yChisUncertainty = np.zeros(yChis.shape)
    # recovering the spectral values
    spectrum = np.dot(yChis, knotsY)
    spectrumVariance = dotVariance(yChis,
                                   knotsY,
                                   yChisUncertainty,
                                   knotsYUncertainty)
    # print(spectrumVariance)
    
    # initialize arrays for holding photon energies and corresponding
    # intensities of stitched spectrum.
    photonEnergies = np.array([])
    intensities = np.array([])
    intensitiesVariance = np.array([])
    # loop over segments to reconstruct spectrum
    for segNum, segment in enumerate(segments):
        energyMin, energyMax = segment
        energyRegs = splineCoordsInv(energyNorms, energyMin, energyMax)
        # appending each segment to form a single array of spectral
        # values. We need to cut the last element from each array as the
        # last element overlaps the first element of the next segment.
        photonEnergies = np.append(photonEnergies, energyRegs[:-1])
        intensities = np.append(intensities, spectrum[:-1, segNum])
        intensitiesVariance = np.append(intensitiesVariance,
                                        spectrumVariance[:-1, segNum])
        if plot:
            plt.plot(energyRegs, spectrum[:, segNum])
            
    if plot:
        plt.ylabel('Spectrum (GW/sr/eV)')
        plt.xlabel('Photon Energy (eV)')
        plt.title('Spectrum from cubic spline knots')
        plt.show()
    return photonEnergies, intensities, intensitiesVariance


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  8 09:25:05 2019

Miscellaneous utilities

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np
import astropy.units as u
import copy

# listing all functions declared in this file so that sphinx-automodapi
# correctly documents them and doesn't document imported functions.
__all__ = ["find_nearest",
           "areDataFramesCompatible",
           ]


# solid angles of channels 1 thru 18 (from C Sorce)
# This is in steradians (added by DHB 3/22/2019)
solidAngles = [4.48e-6,
               2.76e-6,
               2.76e-6,
               2.76e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6,
               7.07e-6]

# radius of OMEGA chamber
chamberRadius = 180 * u.cm

def find_nearest(array, value):
    r"""
    Find nearest value in array and return index, and value as a tuple.
    
    Parameters
    ----------
    array: list, numpy.ndarray
        Array of values to be searched.
        
    value: int, float
        Value for which this function will find the nearest value
        in the array.
    
    Returns
    -------
    idx: int
        Index at which nearest value to input value occurs in the array.
    
    array[idx]: int, float
        The nearest value to the input value.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    
    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx, array[idx]


#TODO Add support for channels of type ndarray
def areDataFramesCompatible(channels, *frames):
    r"""
    Check DataFrame compatibility for specified channels.
    
    Checks if multiple pandas.core.frame.DataFrame objects are compatible and have
    the channels that are requested. Checks that the DataFrames span the same
    energy range. Returns true if the frames pass all checks, false otherwise.
    
    Parameters
    ----------
    channels: list
        List of relevant channels
    
    *frames: pandas.core.frame.DataFrame
        The DataFrames that you want to check for compatiblity with the relevant
        channels
        
    Returns
    -------
    bool
        True if frames are compatible with the requested channels, and False
        otherwise.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    #deep copy so we don't modify the passed channel list
    ch = copy.deepcopy(channels)
    ch.append('Energy(eV)')
    cols = []
    frameShapes = []
    energyCols = []
    for frame in frames:
        cols.append(set(frame.columns.values))
        frameShapes.append(frame.shape)
        #check if the frame has an Energy(eV) col
        if not 'Energy(eV)' in frame.columns:
            #if not stop here, it definitely wont pass the next tests
             return False
        energyCols.append(frame['Energy(eV)'].values)
     
    #check if all frames have 2d shape
    for shape in frameShapes:
        if len(shape) != 2:
            return False
    
    #check if the frames have the channels we need
    overlap = set.intersection(*cols)
    if not set(ch).issubset(overlap):
        return False
    
    #check if the energy columns are equal
    energyColToChangeAgainst = energyCols[0]
    for energyCol in energyCols:
        #check if the energy column lengths are equal
        if len(energyColToChangeAgainst) != len(energyCol):
            return False
        #check if the energy column elements are equal
        if not np.array_equal(energyColToChangeAgainst, energyCol):
            return False
    
    #no more tests to run, DataFrames are compatible with the relevant channels
    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Mar  8 09:23:02 2019

Utilities for working with DANTE response functions (e.g. plotting, locating
edges).

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np


# listing all functions declared in this file so that sphinx-automodapi
# correctly documents them and doesn't document imported functions.
__all__ = ["knotFind"]


def knotFind(channels,
             responseFrame,
             forceKnot=np.array([]),
             knotBoundary=0,
             boundary='y0'):
    r"""Find knot points.
    
    Find knot points for cubic splines based on positions of K-edges
    of each DANTE channel filter.
    
    Parameters
    ----------
    channels: list, numpy.ndarray
        List or array of relevant channels
    
    responseFrame: pandas.core.frame.DataFrame
        Pandas dataFrame containing response functions for each DANTE
        channel. See loadResponses().
        
    forceKnot: numpy.ndarray
        Numpy array where first column is channelNumber and second column
        is the corresponding photonEnergy we wish to force. Use this for
        channels that do not have a distinct K-edge.
        
    knotBoundary: float
        Photon energy value for either y_0 or y_{n+1} boundary condition.
        This value gets appended to the array of photon energies otherwise
        found by knotFind().
        
    Returns 
    -------
    knotsAppend: numpy.ndarray
        An array of knot points, with each element corresponding to a channel
        or boundary condition.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    knots = np.zeros(len(channels))
    for idx, channel in enumerate(channels):
        # if forceKnot isn't an empty list, then we go about forcing the user
        # provided values.
        if forceKnot.size != 0:
            if channel in forceKnot[:,0]:
                # if the channel is forced by the user then put the user provided
                # photon energy into the knots array
                forceIdx = np.where(forceKnot[:,0] == channel)
                knots[idx] = forceKnot[forceIdx, 1]
            else:
                # finding largest negative gradient, which should correpond to the
                # K-edge of the DANTE channel filter.
                grad = -np.gradient(responseFrame[channel])
                maxIndex = np.argmax(grad)
                knots[idx] = responseFrame['Energy(eV)'][maxIndex]
        else:
            # finding largest negative gradient, which should correpond to the
            # K-edge of the DANTE channel filter.
            grad = -np.gradient(responseFrame[channel])
            maxIndex = np.argmax(grad)
            knots[idx] = responseFrame['Energy(eV)'][maxIndex]
        
    if boundary == 'y0':
        knotsAppend = np.append([knotBoundary], knots)
    elif boundary == 'yn+1':
        knotsAppend = np.append(knots, [knotBoundary])
    else:
        raise Exception(f"No method found for boundary {boundary}.")
    return knotsAppend


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tues June 16 13:48:21 2020

Utilities for calculating response uncertainty

@author: Myles Brophy
"""

# python modules
import numpy as np
import xarray as xr
import pandas as pd
import matplotlib.pyplot as plt
import datetime
#custom modules
from fiducia.cspline import splineCoordsInv, segmentsArr, dToyArr, yChiCoeffArrEnergies, responseInterp, detectorArr
from fiducia.misc import areDataFramesCompatible
from fiducia.stats import trapzVariance, gradientVariance, interpVariance
from fiducia.response import knotFind
from fiducia.loader import loadResponses, loadResponseUncertainty
# listing all functions declared in this file so that sphinx-automodapi
# correctly documents them and doesn't document imported functions.
__all__ = ["detectorErrMC",
           "knotVarianceFind",
           "responseInterpVariance",
           "fancyTrapz2Variance",
           "detectorArrVariance",
           "detectorUncertainty"
           ]



    
def detectorErrMC(detArr, detArrVariance, samples=10000,
                  boundary="y0", MChistogram=False):
    r"""
    Monte Carlo simulation and statistics to determine cubic spline uncertainty.
    
    Calculate the cubic spline matrix uncertainty using a Monte Carlo simulation
    and statistics on the MC's output.
    
    Parameters
    ----------
    detArr: numpy.ndarray
        Matrix representing the spectrally integrated folding of the detector
        response with a cubic spline interpolation of the x-ray spectrum. See
        'cspline.detectorArr()'.
    
    detArrVariance:  numpy.ndarray
        A DataFrame containing the uncertainty for each Dante channel for the
        photon energy range that detArr spans.
    
    samples : int, optional
        Number of MCs amples to run. Default is `10000`.
    
    boundary : str, optional
        Choose whether yGuess corresponds to :math:`y_0` (lowest photon energy) or
        :math:`y_{n+1}` (highest photon energy) boundary condition. This should
        correspond to the photon energy value given in knots. Options are `y0`
        or `yn+1`. Default is 'y0'.
    
    MChistogram : bool, optional
        Plot histograms corresponding to each variant of detArr generated with
        Monte Carlo uncertainty propagation.
        Default is `False`.
    
    Returns
    -------
    stdErrorMatrix : numpy.ndarray
         A numpy.ndarray with the standard deviation of the inverted matrices
         generated using random weights based on the channel uncertainty.
        
    Raises
    ------
    Exception
        If `boundary` doesn't equal `y0` or `yn+1`.
    ValueError
        If the shapes of `detArr` and `detUncertaintyArr` aren't equal.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    
    if detArr.shape != detArrVariance.shape:
        raise ValueError(f"Shape of the cubic spline matrix ({detArr.shape}) is not equal to"+
                         f" the shape of the detector uncertainty matrix ({detArrVariance.shape})")
    #square root the uncertainty matrix because it is sigma^2 and we want sigma
    detArrUncertainty = np.sqrt(detArrVariance)
    #make a 3d matrix, where each 2d slice is an error matrix generated by MC
    print("Generating "+str(samples) + " MC samples...")
    error = xr.DataArray(np.zeros((samples,)+detArr.shape), dims=['sample', 'channel', 'knot_point'])
    for s in range(samples):
        error[s] = detArr + np.random.normal(loc=0, scale=1, size=detArr.shape)*detArrUncertainty

    print("Finished generating MC samples.")
    
    detArrInv = xr.DataArray(np.linalg.inv(detArr), dims=['channel', 'knot_point'], attrs={'boundary':boundary})  
    invError = xr.DataArray(np.zeros(error.shape), dims=['sample', 'channel', 'knot_point'])
    
    # #invert each 2d slice of the error matrix
    for sample in range(samples):
        invErrSlice = np.linalg.inv(error[sample])
        invError[sample] = invErrSlice
            
    # #find std of each element compared to other samples
    stdDetArrInv = invError.std(dim="sample")
    stdDetArrInv.attrs['boundary'] = boundary
    
    if MChistogram:
        for ch in invError['channel']:
            for knot in invError['knot_point']:
                # print(len(invError.sel(channel=ch, knot_point=knot).values))
                plt.hist(invError.sel(channel=ch, knot_point=knot).values,
                         samples,
                         alpha=1,
                         label=f'Channel {ch.values} knot {knot.values}')
                plt.axvline(detArrInv.sel(channel=ch, knot_point=knot).values,
                            label='DetArrInv',
                            color='red')
                plt.axvline(detArrInv.sel(channel=ch,knot_point=knot).values + stdDetArrInv.sel(channel=ch, knot_point=knot).values,
                            label='DetArrInv+sigma',
                            color='purple')
                plt.axvline(detArrInv.sel(channel=ch, knot_point=knot).values - stdDetArrInv.sel(channel=ch, knot_point=knot).values,
                            label='DetArrInv-sigma',
                            color='green')
                plt.legend(loc='upper right')
                plt.title("Detector MC")
                plt.show()
                
    return detArrInv, stdDetArrInv
    

#TODO check to see if this has any effect on final error bars. Pawel thinks not.
def knotVarianceFind(channels,
                     responseUncertaintyFrame=None,
                     forceKnot=np.array([]), 
                     knotBoundaryY=1e-77,
                     boundary='y0'):
    r"""
    Modification of response.knotFind()

    Parameters
    ----------
    channels: numpy.ndarray
        Array of DANTE channel numbers. 

    responseUncertaintyFrame: pandas.core.frame.DataFrame, optional
        DataFrame holding percent uncertainties of DANTE channel responses as
        a function of photon energy (not normalized). The default is `None`.

    forceKnot : TYPE, optional
        DESCRIPTION. The default is `np.array([])`.
    
    knotBoundaryY : float, optional
        Guess for position of y_0 or y_{n+1} knot point. Default is 1e-77.
    
    boundary: str, optional
        Choose whether yGuess corresponds to :math:`y_0` (lowest photon
        energy) or :math:`y_{n+1}` (highest photon energy) boundary condition.
        This should correspond to the photon energy value given in knots.
        Options are `y0` or `yn+1`. Default `y0`.

    Returns
    -------
     knotUncertainty: numpy.ndarray
         An array of uncertainty in knot points, with each element
         corresponding to a channel or boundary condition. 
         See :func:`response.knotFind`.
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    #modified from response.knotFind
    knotsUncertainty = np.zeros(len(channels))
    
    #check if uncertainty frame is passed.
    if responseUncertaintyFrame is None:
        #return ndarray of length channels + 1 for boundary condition
        return  np.zeros(len(channels)+1)
    
    for idx, channel in enumerate(channels):
        # if forceKnot isn't an empty list, then we go about forcing the user
        # provided values.
        if forceKnot.size != 0:
            if channel in forceKnot[:,0]:
                # if the channel is forced by the user then put the user provided
                # photon energy into the knots array
                forceIdx = np.where(forceKnot[:,0] == channel)
                knotsUncertainty[idx] = forceKnot[forceIdx, 1]
            else:
                # finding largest negative gradient, which should correpond to the
                # K-edge of the DANTE channel filter.
                grad = -gradientVariance(responseUncertaintyFrame[channel])
                maxIndex = np.argmax(grad)
                knotsUncertainty[idx] = responseUncertaintyFrame['Energy(eV)'][maxIndex]
        else:
            # finding largest negative gradient, which should correpond to the
            # K-edge of the DANTE channel filter.
            grad = -gradientVariance(responseUncertaintyFrame[channel])
            maxIndex = np.argmax(grad)
            knotsUncertainty[idx] = responseUncertaintyFrame['Energy(eV)'][maxIndex]
        
    if boundary == 'y0':
        knotsUncertaintyAppend = np.append([knotBoundaryY], knotsUncertainty)
    elif boundary == 'yn+1':
        knotsUncertaintyAppend = np.append(knotsUncertainty, [knotBoundaryY])
    else:
        raise Exception(f"No method found for boundary {boundary}.")
    return knotsUncertaintyAppend


def responseInterpVariance(energyNorm,
                           energyMin,
                           energyMax,
                           responseUncertaintyFrame,
                           channels):
    r"""
    Given a DANTE detector response as a function of energy, convert the
    response to normalized photon energy, t, over a given spline segment, and
    return interpolated response values for a given value of t. Returns an
    array of interpolated responses corresponding to the number of channels.
    
    Parameters
    ----------
    energyNorm: float, numpy.ndarray
        normalized photon energy
        
    energyMin: float
        Lower bound photon energy of the spline segment over which we are
        normalizing.
        
    eneryMax: float
        Upper bound photon energy of the spline segment over which we are
        normalizing.
        
    responseFrame: pandas.core.frame.DataFrame
        DANTE channel responses as a function of photon energy (not normalized).

    channels: numpy.ndarray
        numpy array of DANTE channel numbers.
    
    Returns
    -------
    responsesInterpdVariance: numpy.ndarray
        Returns a matrix of (energyNorms, channels) of response functions.
    
    Notes
    -----
    
    See also
    --------
    cspline.repsonseInterp
    
    Examples
    --------

    """
    chLen = len(channels)
    if np.shape(energyNorm):
        # if energyNorm is a vector, then get the length of that vector
        energyLen = len(energyNorm)
    else:
        # otherwise the length is just 1
        energyLen = 1
    # construct an array of interpolated response functions based on
    # number of photon energy points we want and number of DANTE channels
    responsesInterpdVariance = np.zeros((energyLen, chLen))
    # fetch the original array of energy values (not normalized)
    energyArr = responseUncertaintyFrame['Energy(eV)']
    # converting normalized energy values to un-normalized energy values,
    # since our response functions are in terms of absolute photon energy
    energyReg = splineCoordsInv(energyNorm, energyMin, energyMax)
    for idx, channel in enumerate(channels):
        responseUncertaintyArr = responseUncertaintyFrame[channel]
        responsesInterpdVariance[:,idx] = interpVariance(x=energyReg,
                                                         xp=energyArr,
                                                         fpUnc=responseUncertaintyArr)
    return responsesInterpdVariance


def fancyTrapz2Variance(energyNorms,
                        yChis,
                        segments,
                        responseUncertaintyFrame,
                        channels,
                        interpProp=True):
    r"""
    Calculate the variance when propogating uncertainties
    through :func:`fiducia.cspline.fancyTrapz2`.

    Parameters
    ----------
    energyNorms : numpy.ndarray
        Array of normalized energies over which the integral is computed.
    
    yChis : numpy.ndarray
        3D array corresponding to the :math:`M_{y \chi}` coefficients.
        Array shape corresponds to (`energyNorms`, `chLen`, `dToY`). 
        See :func:`fiducia.error.detectorArrVariance`
    
    segments : numpy.ndarray
        Array of segments produced by :func:`segmentsArr` with the knots
   
    responseUncertaintyFrame : pandas.core.frame.DataFrame
        DataFrame holding uncertainty percentages of DANTE channel responses
        as a function of photon energy (not normalized).
        
    channels : numpy.ndarray
        Array of DANTE channel numbers.
        
    interpProp : bool, optional
        Boolean to decide if :func:`error.responseInterpVariance` should be
        used. If `False, :func:`cspline.responseInterp()` is used, speeding
        up the calculation. Note that the uncertainty is would not be
        propagated correctly if `False`. With future optimizations, this
        option to choose may be removed. Default is `True`.
    
    Returns
    -------
    integArrVariance : xarray.Dat
        A matrix containing the folded integration of the :math:`M_{y \chi}` 
        matrix and response function uncertainty matrix, with respect to
        normalized photon  energy.
        Has shape (`len(channels)`, `len(segments)`, `len(knotIndex)`).
        
    Notes
    -----
    
    See also
    --------
    cspline.fancyTrapz2()
    
    Examples
    --------
        
    """
    shape = np.shape(yChis)
    segmentsLen = np.shape(segments)[0]
    knotsLen = shape[2]
    chLen = segmentsLen
    # initialize integArr for storing photon energy integrated values
    integArrVariance = xr.DataArray(np.zeros((chLen, segmentsLen, knotsLen)),
                                    dims=['channel', 'segment', 'knot_point'],
                                    coords={'channel':channels})  

    # loop over relevant DANTE channels for analysis
    for channelIdx in np.arange(chLen):
        # loop over photon energy segments (between knot points)
        for segmentNum, segment in enumerate(segments):
            energyMin, energyMax = segment
            # loop over knot points
            for knotNum in np.arange(knotsLen):
                # multiplication of response by spline matrix
                #use interpolation propagation by default, but runs slower
                #might not have much an effect. For now we'll let the user decide
                #need to optimize reponseInterpVariance
                if interpProp:
                    responsesVariance = responseInterpVariance(energyNorms,
                                                               energyMin,
                                                               energyMax,
                                                               responseUncertaintyFrame,
                                                               channels)
                    responsesUncertainty = np.sqrt(responsesVariance)
                else:
                    responsesVariance = responseInterp(energyNorms,
                                                       energyMin,
                                                       energyMax,
                                                       responseUncertaintyFrame,
                                                       channels)
                    
                multArr = np.abs(yChis[:, segmentNum, knotNum] * responsesUncertainty[:, channelIdx])
                integVar = (energyMax - energyMin) ** 2 * trapzVariance(multArr,
                                                                        x=energyNorms)
                integArrVariance[channelIdx, segmentNum, knotNum] = integVar
    return integArrVariance


def detectorArrVariance(channels, knots, responseUncertaintyFrame, boundary="y0", npts=1000):
    r"""
    Propagates uncertanity through :func:`cspline.detectorArr() to find the variance in :math:`M_{int}`.
    
    Parameters
    ----------
    channels : numpy.ndarray
        Array of DANTE channel numbers.
    
    knots : numpy.ndarray
        Array of photon energies describing positions of spline knots.
    
    responseUncertaintyFrame : pandas.core.frame.DataFrame
        DataFrame holding uncertainty percentages of DANTE channel responses
        as a function of photon energy (not normalized).
        
    npts : int, optional
        Number of points used in computing the integral. The default is 1000.
        
    Returns
    -------
    detArrVariance : xarray.DataArray
        2D array of channels and knot points uncertainties of shape `(n, n+1)`.

    detArrVarianceBoundaryCol : xarray.DataArray
        Column of variances in the cublic spline matrix corresponding to the 
        knots at the boundary chosen with `boundary`.
        
    Notes
    -----
    Covariances between segments is not currently accounted for. This
    covariance should be small compared to the other uncertainties, but should
    be noted.
    
    See also
    --------
    cspline.detectorArr()
    
    Examples
    --------
    
    """
    # number of DANTE channels where we have useful measurements
    chLen = len(channels)
    # initialize normalized energies array
    # array of normalized energies over which we do the integral
    energyNorms = np.linspace(0, 1, num=npts)
    # producing segments from knotsUncertainty. Use the same segmentsArr
    #because there is no calculation done
    segments = segmentsArr(knots)
    # calculating array for converting from values of D_i to y_i. This
    # is an optimization as this array is constant!
    dToY = dToyArr(chLen)
    # M_{y \chi} coefficients array corresponding to given normalized
    # energies. Array shape is (energyNorms, segments, knotIndex).
    yChis = yChiCoeffArrEnergies(energyNorms, chLen, dToY)
    print("running fancytrapz2Variance")
    integFoldArrVariance = fancyTrapz2Variance(energyNorms,
                               yChis,
                               segments,
                               responseUncertaintyFrame,
                               channels)
    print("finished fancytrapz2")
    # sum along segment axis, as each segment must contribute to the
    # overall signal.
    #detArrVariance = np.sum(integFoldArrVariance, axis=1)
    detArrVariance = integFoldArrVariance.sum(dim="segment")
    detArrVariance.attrs['boundary'] = boundary
    if boundary == "y0":
        # extracting column corresponding to y0
        detArrVarianceBoundaryCol = detArrVariance.isel(knot_point=0)
        detArrVariance = detArrVariance.isel(knot_point=slice(1, None))
    elif boundary == "yn+1":
        # extracting column corresponding to y_{n+1}
        detArrVarianceBoundaryCol = detArrVariance.isel(knot_point=-1)
        detArrVariance = detArrVariance.isel(knot_point=slice(None, -1))
    else:
        raise Exception(f"No method found for boundary {boundary}.")
    return detArrVariance, detArrVarianceBoundaryCol


def detectorUncertainty(channels,
                        responseFile,
                        responseUncertaintyFile=None,
                        boundary="y0",
                        npts=1000,
                        samples=1000,
                        MChistogram=False,
                        saveDataset=True,
                        csplineDatasetFile=''):
    r"""
    Finds the cspline detector matrix, it`s inverse matrix and std matrix
    using Monte Carlo uncertainty propagation.

    Propagates response uncertainties through 

    Parameters
    ----------
    channels: numpy.ndarray
        Array of DANTE channel numbers.
        
    responseFile: str
        Path to the `.csv` holding DANTE channel responses as a function of 
        photon energy (not normalized).
    
    responseUncertaintyFile: str, optional
        Path to the `.csv` holding DANTE channel response uncertainties as a 
        function of photon energy. Uncertainty values provided as percentages.

    boundary: str, optional
        Choose whether yGuess corresponds to :math:`y_0` (lowest photon
        energy) or :math:`y_{n+1}` (highest photon energy) boundary condition.
        This should correspond to the photon energy value given in knots.
        Options are `y0` or `yn+1`. Default 'y0'.
    
    npts: int, optional
        Number of points used in computing the integral. Default is 1000.
    
    samples: int, optional
        Number of samples to generate during Monte Carlo propagation.
        See :func:`error.detectorErrMC`. Default is 1000.
        
    Returns
    -------
    detArr : xarray.DataArray
        Matrix representing the spectrally integrated folding of the detector
        response with a cubic spline interpolation of the x-ray spectrum.
        2D array of channels and knot points of shape (n, n).
 
    detArrBoundaryCol : xarray.DataArray
        Column of cublic spline matrix corresponding to the knots at the
        boundary chosen with `boundary`.
        
    detArrVarianceBoundaryCol: xarray.DataArray
        Column of variances in the cublic spline matrix corresponding to the 
        knots at the boundary chosen with `boundary`.
    
    detArrInv : xarray.DataArray
        Inversion of detArr, with the column corresponding to boundary removed
        so detArr is invertible.
   
    stdDetArrInv : xarray.DataArray
        Array of the standard deviation of each element in detArrInv based on
        variance using the `responseUncertaintyFrame` propagated with Monte
        Carlo. 
        
    Notes
    -----
    
    See also
    --------
    
    Examples
    --------
    """
    #load response functions
    responseFrame = loadResponses(channels, responseFile)
    
     #create responseUncertainty with 0s if not given.
    if responseUncertaintyFile is None:
        responseUncertaintyFrame = responseFrame.copy()
        cols = list(responseFrame.columns.values)
        cols.remove('Energy(eV)')
        for col in cols:
            responseUncertaintyFrame.loc[:,col] = 0
    else:
        responseUncertaintyFrame = loadResponseUncertainty(responseFrame, responseUncertaintyFile)

    knots = knotFind(channels, responseFrame)
    
    #knotsVariance = knotVarianceFind(channels, responseUncertaintyFrame)
    if not areDataFramesCompatible(channels, responseFrame, responseUncertaintyFrame):
        raise ValueError("Response frame and response uncertainty frame are not compatible. Check formats.")
    detArr, detArrBoundaryCol = detectorArr(channels,
                                            knots,
                                            responseFrame,
                                            boundary,
                                            npts)
    detArrVariance, detArrVarianceBoundaryCol = detectorArrVariance(channels,
                                                                    knots,
                                                                    responseUncertaintyFrame,
                                                                    boundary,
                                                                    npts)
    detArrInv, stdDetArrInv = detectorErrMC(detArr,
                                            detArrVariance,
                                            samples,
                                            boundary,
                                            MChistogram)
    #combine into Dataset
    csplineDataset = xr.Dataset(data_vars={'detArr' : detArr,
                                           'detArrBoundaryCol' : detArrBoundaryCol,
                                           'detArrVarianceBoundaryCol' : detArrVarianceBoundaryCol,
                                           'detArrInv' : detArrInv,
                                           'stdDetArrInv' : stdDetArrInv}, 
                                attrs={'boundary' : boundary,
                                       'channels' : np.asarray(channels), 
                                       'yGuess' : 0,
                                       'responseFile' : responseFile, 
                                       'responseUncertaintyFile' : responseUncertaintyFile,
                                       'generatedDatetime' : str(datetime.datetime.now())})
    if saveDataset:
        #save as netCDF as recommended by http://xarray.pydata.org/en/stable/io.html#netcdf
        if not csplineDatasetFile:
            csplineDatasetFile = 'csplineDataset_' + str(datetime.datetime.now().date()) + '.nc'
        csplineDataset.to_netcdf(csplineDatasetFile)
    return csplineDataset 



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Aug 21 19:17:24 2020

A couple of tests to ensure that fixtures from conftest.py are loading
correctly for the pytest sessions, and that the returned data are
behaving as expected.

@author: Pawel M. Kozlowski
"""

import pytest

def test_info_fixture(example_info):
    """Testing whether conftest.py fixture is working as expected."""
    assert example_info["Angle"] == 69.09
    

def test_data_fixture(load_example_data):
    """Testing whether conftest.py fixture is working as expected."""
    expectedChOn = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 17]
    testTrue = load_example_data["On Channels"] == expectedChOn
    errStr = (f"Header loaded incorrectly. Expected on channels are "
              f"{expectedChOn}, but got {load_example_data['On Channels']} "
              "instead.")
    assert testTrue, errStr